/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "rollball.hpp"
#include "fggl/gfx/camera.hpp"
#include "fggl/gfx/phong.hpp"

#include "fggl/input/camera_input.hpp"

#include "fggl/entity/entity.hpp"
#include "fggl/entity/loader/loader.hpp"

#include "fggl/debug/draw.hpp"

#include <array>

static const fggl::util::GUID playerPrefab = "rb_player"_fid;
static const fggl::util::GUID collectablePrefab = "rb_collectable"_fid;
static const fggl::util::GUID WallNPrefab = "rb_wallX"_fid;
static const fggl::util::GUID WallEPrefab = "rb_wallZ"_fid;
static const fggl::util::GUID floorPrefab = "rb_floor"_fid;

static void setup_camera(fggl::entity::EntityManager& world) {
	auto prototype = world.create();

	// setup camera position/transform
	auto& transform = world.add<fggl::math::Transform>(prototype);
	transform.origin(glm::vec3(10.0f, 3.0f, 10.0f));

	// setup camera components
	world.add<fggl::gfx::Camera>(prototype);
}

static fggl::entity::EntityID setup_environment(fggl::entity::EntityManager& world, fggl::entity::EntityFactory* factory, const fggl::math::vec2& size, demo::RollState& state) {
	// ensure the state is clean
	state.closestPickup = fggl::entity::INVALID;
	state.mode = demo::DebugMode::NORMAL;
	state.time = 0.0F;
	state.player = world.findByName("player");
	state.collectables = world.findByTag("collectable");

	return state.player;
}


namespace demo {

	constexpr fggl::math::vec2 WORLD_SIZE{40.0F, 40.0F };
	constexpr fggl::math::vec3 COLLECTABLE_ROTATION{15, 30, 45};
	constexpr fggl::math::vec3 COLOUR_WHITE{1.0F, 1.0F, 1.0F};
	constexpr fggl::math::vec3 COLOUR_BLUE{0.0F, 0.0F, 1.0F};
	constexpr float MOVE_FORCE{3.0F};

	RollBall::RollBall(fggl::App &app) : Game(app), m_phys(nullptr) {
	}

	void RollBall::activate() {
		Game::activate();
		fggl::debug::log(fggl::debug::Level::info, "RollBall::activate()");

		// attach physics
		auto* physService = m_owner.service<fggl::phys::PhysicsProvider>();
		auto* entFactory = m_owner.service<fggl::entity::EntityFactory>();
		m_phys = physService->create(&world(), entFactory);

		auto* scriptProvider = m_owner.service<fggl::script::ScriptProvider>();

		m_scripts = scriptProvider->create();
		m_scripts->setGlobal("state", this);

		// asset loader
		auto* assetLoader = m_owner.service<fggl::assets::Loader>();
		assetLoader->load("rollball.yml", fggl::entity::ENTITY_PROTOTYPE, entFactory);
		assetLoader->loadChain( fggl::assets::make_asset_id_rt("core", "rollerball/rollball.yml"), this);

		// collectable callbacks
		/*auto* collectableCallbacks = world().get<fggl::phys::CollisionCallbacks>(prefabs.collectable);
		collectableCallbacks->onEnter = [this](auto ourEntity, auto theirEntity) {
			if ( theirEntity == state.player) {
				//if ( ourEntity == state.closestPickup ) {
				//	// we're the closest pickup and we're about to get killed, so need to not be.
				//	state.closestPickup = fggl::ecs::NULL_ENTITY;
				//}
				world().destroy(ourEntity);
			}
		};*/

		// actual scene objects
		setup_camera(world());

		// create a 20x20 grid
		setup_environment(world(), m_owner.service<fggl::entity::EntityFactory>(), WORLD_SIZE, state);

		// activate scripts
		if ( m_scripts != nullptr ) {
			m_scripts->onActivate();
			m_scripts->load("rollball.lua");
		}
	}

	void RollBall::deactivate() {
		// deactivate scripts
		if ( m_scripts != nullptr ) {
			m_scripts->onActivate();
		}

		// we need to clean up physics
		if ( m_phys != nullptr ) {
			delete m_phys;
			m_phys = nullptr;
		}

		// now let the rest of the scene do its bit
		Game::deactivate();
	}

	fggl::math::vec3 calc_move_vector(const fggl::input::Input& input) {
		constexpr fggl::math::vec3 forward = fggl::math::FORWARD;
		constexpr fggl::math::vec3 right = fggl::math::RIGHT;
		fggl::math::vec3 force = fggl::math::VEC3_ZERO;

		if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_W))) {
			force += forward;
		}

		if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_S))) {
			force -= forward;
		}

		if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_A))) {
			force -= right;
		}

		if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_D))) {
			force += right;
		}

		return force;
	}

	void RollBall::update(float deltaTime) {
		Game::update(deltaTime);
		m_phys->step();

		auto& input = this->input();

			if ( state.player != fggl::entity::INVALID ) {
				auto &world = this->world();

				// mode selection
				if (input.keyboard.pressed(glfwGetKeyScancode(GLFW_KEY_SPACE))) {
					state.mode = (DebugMode)( ((int)state.mode + 1) % 3 );
				}

				// setup dynamic force
				auto force = calc_move_vector(input);
				if ( force != fggl::math::VEC3_ZERO ) {
					force = glm::normalize(force) * MOVE_FORCE;
					auto& dynamics = world.get<fggl::phys::Dynamics>(state.player);
					dynamics.force += force;
				}

				// track player position with camera
				{
					auto cameras = world.find<fggl::gfx::Camera>();
					auto cam = cameras[0];
					auto& camComp = world.get<fggl::gfx::Camera>(cam);
					camComp.target = world.get<fggl::math::Transform>(state.player).origin();

					auto& camTransform = world.get<fggl::math::Transform>(cam);
					camTransform.origin( camComp.target + cameraOffset );
				}

				// distance mode
				if ( state.mode == DebugMode::DISTANCE ) {
					closestPickup(world);
				} else {
					if ( state.closestPickup != fggl::entity::INVALID ) {
						if ( world.alive(state.closestPickup) ){
							auto *renderer = world.tryGet<fggl::gfx::PhongMaterial>(state.closestPickup);
							if (renderer != nullptr) {
								renderer->diffuse = COLOUR_WHITE;
							}
						}

						state.closestPickup = fggl::entity::INVALID;
					}
				}

				// make the cubes spin
				spinCubes(world, deltaTime);
				state.time += deltaTime;
		}

		// tick scripts
		if ( m_scripts != nullptr ) {
			m_scripts->onUpdate();
		}
	}

	void RollBall::closestPickup(fggl::entity::EntityManager &world) {
		float closestDistance = FLT_MAX;
		auto closestEntity = fggl::entity::INVALID;
		auto playerPos = world.get<fggl::math::Transform>(state.player).origin();

		for ( const auto& pickup : state.collectables ) {
			if ( !world.alive(pickup) ) {
				continue;
			}

			auto& transform = world.get<fggl::math::Transform>(pickup);
			auto distance = glm::distance(transform.origin(), playerPos);
			if ( distance < closestDistance) {
				closestDistance = distance;
				closestEntity = pickup;
			}
		}

		if ( state.closestPickup != closestEntity ) {
			if ( world.alive(state.closestPickup) ){
				// deal with the previous closest pickup
				auto *renderer = world.tryGet<fggl::gfx::PhongMaterial>(state.closestPickup);
				if (renderer != nullptr) {
					renderer->diffuse = COLOUR_WHITE;
				}
			}

			// now, deal with the new closest
			state.closestPickup = closestEntity;
			if (closestEntity != fggl::entity::INVALID) {
				auto *renderer = world.tryGet<fggl::gfx::PhongMaterial>(state.closestPickup);
				if (renderer != nullptr) {
					renderer->diffuse = COLOUR_BLUE;
				}
			}
		}

		// closest pickup should face the player
		if ( state.closestPickup != fggl::entity::INVALID) {
			auto& transform = world.get<fggl::math::Transform>(state.closestPickup);
			auto& playerTransform = world.get<fggl::math::Transform>( state.player );
			transform.lookAt(playerTransform.origin());

			// lazy, so using the debug draw line for this bit
			dd::line( &playerTransform.origin()[0], &transform.origin()[0], dd::colors::White );
		}

	}

	void RollBall::spinCubes(fggl::entity::EntityManager& world, float deltaTime) {
		// rotation
		for ( const auto& entity : state.collectables ) {
			if ( !world.alive(entity) || entity == state.closestPickup ) {
				continue;
			}

			// rotate the cubes
			auto& transform = world.get<fggl::math::Transform>(entity);
			transform.rotateEuler( COLLECTABLE_ROTATION * deltaTime );
		}

	}

	void RollBall::render(fggl::gfx::Graphics &gfx) {
		Game::render(gfx);
	}

} // namespace demo

